/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License aboolean with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

/**
 * <p>Abstract array implementation.</p>
 *
 * @author Leon van Zantvoort
 */
abstract class AbstractArray<E> implements Array<E> {

    /** Start index of range. */
    protected final int offset;

    /** Length of range. */
    protected int length;

    /** Array characteristics. */
    protected int characteristics;

    /**
     * Package visible constructor, to be used by sub classes only.
     *
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     */
    AbstractArray(int offset, int length, int characteristics) {
        if (offset < 0 || length < 0) {
            throw new IndexOutOfBoundsException();
        }
        this.offset = offset;
        this.length = length;
        this.characteristics = filterCharacteristics(characteristics) | (length > 1 ? 0 : SORTED | DISTINCT);
    }

    /**
     * Checks if the given index is in range. If not, throws an appropriate runtime exception.
     *
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    protected final void rangeCheck(int index) {
        if (index >= length || index < 0) {
            throw new IndexOutOfBoundsException(outOfBoundsMsg(length, index));
        }
    }

    /**
     * A version of rangeCheck used by add and addAll.
     *
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    protected final void rangeCheckForAdd(int index) {
        if (index > length || index < 0) {
            throw new IndexOutOfBoundsException(outOfBoundsMsg(length, index));
        }
    }

    /**
     * Checks if the given boundary is within this array's boundaries.
     *
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    protected final void subArrayCheck(int offset, int length) {
        if (offset < this.offset || offset + length > this.offset + this.length || length < 0) {
            throw new IndexOutOfBoundsException(outOfBoundsMsg(offset, length));
        }
    }


    /**
     * Constructs an IndexOutOfBoundsException detail message.
     * Of the many possible refactorings of the error handling code,
     * this "outlining" performs best with both server and client VMs.
     */
    protected final String outOfBoundsMsg(int offset, int length) {
        return String.format("Offset: %d, Length: %d, SubArray Offset: %d, SubArray Length: %d", this.offset, this.length, offset, length);
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    protected abstract int filterCharacteristics(int characteristics);

    /**
     * Returns the characteristics set for this array. The result is represented as ORed values from
     * {@link #NONNULL}, {@link #MUTABLE}, {@link #UNBOUNDED}, {@link #SORTED}, {@link #DISTINCT}.
     * Modification of the array contents may result in updated characteristics.
     *
     * @return the characteristics set for this array.
     */
    @Override
    public final int characteristics() {
        return characteristics;
    }

    /**
     * <p>Returns the offset relative to the backing array.</p>
     *
     * @return offset relative to the backing array.
     */
    @Override
    public final int offset() {
        return offset;
    }

    /**
     * <p>Returns the number of elements of this array. All index related methods must respect the boundary represented
     * by this value.</p>
     *
     * <p>Note that for mutable arrays, the backing array - as returned by {@code MutableArray.array()} -
     * may be larger.</p>
     *
     * @return length of array.
     */
    @Override
    public final int length() {
        return length;
    }

    /**
     * Returns {@code true} if this array contains no elements.
     *
     * @return {@code true} if this array contains no elements.
     */
    @Override
    public final boolean isEmpty() {
        return length == 0;
    }
}

