/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.nio.DoubleBuffer;

/**
 * <p>Capacity bounded array of double primitives.</p>
 *
 * <p>Bounded double arrays are constructed by the static interface methods of this interface. The {@code newInstance}
 * method returns a new empty array. The {@code copyOf} method returns a new array representing a deep copy of the
 * array argument. The {@code valueOf} methods wrap a new {@code BoundedDoubleArray} around a native array or buffer.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Bounded primitive arrays, such as {@code BoundedDoubleArray}, always report {@code BOUNDED}, {@code MUTABLE} and
 * {@code NONNULL}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * @author Leon van Zantvoort
 */
public interface BoundedDoubleArray extends DynamicDoubleArray, BoundedArray<Double> {

    /**
     * Creates an empty array with the specified capacity.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code BOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param capacity the max capacity of the array.
     * @throws IllegalArgumentException if the specified capacity is negative.
     * @return a new bounded array.
     */
    static BoundedDoubleArray newInstance(int capacity) {
        if (capacity < 0) {
            throw new IllegalArgumentException("Illegal capacity: " + capacity);
        }
        return valueOf(new double[capacity], 0, 0);
    }

    /**
     * Creates capacity bounded copy of the specified {@code array}, backed by a native array. If the specified
     * {@code array} is mutable, changes to the mutable array are not reflected by this bounded copy.
     *
     * <p>The specified {@code array} can be any array of primitive or object numbers. Reconstructing a copy of this
     * array may involve rounding or as implemented by {@link Number#doubleValue}. The specified {@code array}
     * must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return a bounded array of double values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static BoundedDoubleArray copyOf(@Nonnull Array<? extends Number> array) {
        double[] backingArray = NativeArrays.doubleArray(array);
        if (backingArray != null) {
            return Arrays.newBoundedArray(backingArray, array.offset(), array.length(), array.characteristics(), true);
        }
        return Arrays.newBoundedArray(NativeArrays.toDoubleArray(array), 0, array.length(), array.characteristics(), false);
    }

    /**
     * Creates an bounded array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return an bounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static BoundedDoubleArray valueOf(@Nonnull double... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates an bounded array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return an bounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static BoundedDoubleArray valueOf(@Nonnull double[] array, int offset, int length) {
        return Arrays.newBoundedArray(array, offset, length, 0, false);
    }

    /**
     * <p>Returns an bounded array backed by the specified {@code buffer}. All operations on the resulting array are
     * delegated to a duplicate of the buffer.</p>
     *
     * <p>The contents of {@code buffer} must not be modified directly, or indirectly, after this method returns as this
     * would violate this array's characteristics contract.</p>
     *
     * <p>The returned array is serializable. The deserialized array will allocate a direct buffer if the specified
     * buffer is direct.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param buffer the buffer by which the array will be backed.
     * @return an immutable array backed by the specified buffer.
     * @throws NullPointerException if buffer is null.
     * @throws IllegalArgumentException if buffer is read only.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static BoundedDoubleArray valueOf(@Nonnull DoubleBuffer buffer) {
        return Arrays.newBoundedArray(buffer);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    BoundedDoubleArray offset(int offset);    // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    MutableDoubleArray length(int length);     // Should not return bounded array!

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    MutableDoubleArray subArray(int fromIndex, int toIndex);   // Should not return bounded array!
}
