/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.function.Consumer;

/**
 * <p>Capacity bounded array of object references.</p>
 *
 * <p>This bounded object array is constructed by the static interface methods of {@code BoundedObjectArray}. The
 * {@code newInstance} method returns a new empty array. The {@code copyOf} method returns a new array representing a
 * shallow copy of the array argument. The {@code valueOf} methods wrap a new {@code BoundedObjectArray} around a
 * native array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Bounded primitive arrays, such as this class, always report {@code BOUNDED} and {@code MUTABLE}.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that the array implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. Characteristics are always invalidated if
 * conditions of these characteristics are broken as result of array modification. Characteristics are only added in
 * scenarios in which their conditions can be 'tested' without traversal of elements. For example, the {@code SORTED}
 * characteristic is added if the length of an array is reduced to zero or one element, or if all elements are replaced
 * by an array that is marked as being sorted. The semantics of characteristics processing is documented for each
 * "destructive" method. Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * <p>This class eliminates obsolete references by default to allow the garbage collector to reclaim memory. This is
 * done by setting the elements in 'reserved space' to {@code null}. This feature can be disabled by calling
 * {@code setClearReservedElements(false)}.</p>
 *
 * @author Leon van Zantvoort
 */
@NotThreadSafe
class BoundedObjectArrayImpl<E> extends DynamicObjectArrayImpl<E> implements BoundedObjectArray<E> {

    /**
     * Package visible constructor used to wrap specified {@code array}.
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param characteristics array characteristics.
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> is used.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    BoundedObjectArrayImpl(@Nonnull E[] array, int offset, int length, int characteristics,
                           @Nullable Comparator<? super E> comparator, boolean copy) {
        super(array, offset, length, characteristics, comparator, copy);
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return (characteristics | MUTABLE | BOUNDED) & ~UNBOUNDED;
    }

    /**
     * Sub arrays share serialization proxy with parent arrays intentionally. As a result, deserialized sub arrays become
     * detached from their originating parent.
     */
    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SubBoundedObjectArrayImpl<E> extends BoundedObjectArrayImpl<E> {

        final BoundedObjectArrayImpl<E> parent;

        @SuppressWarnings("unchecked")
        SubBoundedObjectArrayImpl(@Nonnull BoundedObjectArrayImpl<E> parent, int offset, int length) {
            super((E[]) parent.elements, offset, length, parent.characteristics, parent.comparator, false);
            this.parent = parent;
            this.modCount = parent.modCount;
        }

        @Override
        public void setClearReservedElements(boolean flag) {
            parent.setClearReservedElements(flag);  // super.setClearReservedElements() must not be called.
        }

        @Override
        protected boolean clearReservedElements() {
            return parent.clearReservedElements();
        }

        @Override
        protected void checkForComodification() {
            if (modCount != parent.modCount) {
                throw new ConcurrentModificationException();
            }
            parent.checkForComodification();
        }

        @Override
        protected ArrayContext<E> getArrayContext() {
            return parent.inAction ? null : super.getArrayContext();
        }

        @Override
        protected void updateLength(int diff) {
            super.updateLength(diff);
            parent.updateLength(diff);
        }

        @Override
        protected void updateCharacteristics(int elementIndex, E value) {
            super.updateCharacteristics(elementIndex, value);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, value);
            }
        }

        @Override
        protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends E> array) {
            super.updateCharacteristics(elementIndex, array);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, array);
            }
        }

        @Override
        public int characteristics(int characteristics) {
            // Parent array retains characteristics that are shared with sub array. Additional characteristics
            // assigned to sub array are not set for parent array, as it is not guaranteed that these
            // characteristics hold for parent array.
            parent.retainCharacteristics(characteristics);
            return super.characteristics(characteristics);
        }

        @Override
        public ArrayIterator<E> iterator() {
            checkForComodification();
            return super.iterator();
        }

        @Override
        public ArrayIterator<E> iterator(int index) {
            checkForComodification();
            return super.iterator(index);
        }

        @Override
        public void forEach(@Nonnull Consumer<? super E> consumer) {
            checkForComodification();
            super.forEach(consumer);
        }

        @Override
        public E get(int index) {
            checkForComodification();
            return super.get(index);
        }

        @Override
        public @Nonnull E getNonnull(int index) {
             checkForComodification();
             return super.getNonnull(index);
        }

        @Override
        public E set(int index, E value) {
            checkForComodification();
            return super.set(index, value);
        }

        @Override
        public void setAll(@Nonnull Array<? extends E> array) {
            checkForComodification();
            super.setAll(array);
        }

        @Override
        public void add(E object) {
            checkForComodification();
            super.add(object);
        }

        @Override
        public void addAll(@Nonnull Array<? extends E> array) {
            checkForComodification();
            super.addAll(array);
        }

        @Override
        public E remove() {
            checkForComodification();
            return super.remove();
        }

        @Override
        public void clear() {
            checkForComodification();
            super.clear();
        }

        @Override
        public Object[] toArray() {
            checkForComodification();
            return super.toArray();
        }

        @Override
        public <T> T[] toArray(@Nonnull T[] array) {
            checkForComodification();
            return super.toArray(array);
        }

        @Override
        public <T> void toArray(@Nonnull T[] array, int offset, int length) {
            checkForComodification();
            super.toArray(array, offset, length);
        }

        // All default implementations (defender methods) will be checked for concurrent modification either through
        // methods listed above, or through backing array access below.

        @Override
        public Object[] array() {
            checkForComodification();
            return super.array();
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public BoundedObjectArray<E> offset(int offset) {
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new SubBoundedObjectArrayImpl<>(this, newOffset, newLength);
    }
          
    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public MutableObjectArray<E> length(int length) {
        subArrayCheck(offset, length);
        return new SubBoundedObjectArrayImpl<>(this, offset, length);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public MutableObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        subArrayCheck(newOffset, newLength);
        return new SubBoundedObjectArrayImpl<>(this, newOffset, newLength);
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    @Override
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = -3583023241490787519L;

        transient int offset;
        final int length;
        final int capacity;
        final int characteristics;
        final Comparator<? super E> comparator;
        final Class<?> cls;
        final boolean clearReservedElements;
        transient Object[] elements;

        SerializationProxy(BoundedObjectArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.capacity = array.elements.length - offset;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.elements;
            this.clearReservedElements = array.clearReservedElements();
            this.cls = this.elements.getClass().getComponentType();
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int index = offset, toIndex = index + length; index < toIndex; index++) {
                s.writeObject(elements[index]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (Object[]) java.lang.reflect.Array.newInstance(cls, capacity);
            for (int index = 0; index < length; index++) {
                elements[index] = s.readObject();
            }
        }

        @SuppressWarnings("unchecked")
        private Object readResolve() {
            BoundedObjectArrayImpl<E> array = new BoundedObjectArrayImpl<>((E[]) elements, 0, length, characteristics, comparator, false);
            array.setClearReservedElements(clearReservedElements);
            return array;
        }
    }
}
