/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.nio.DoubleBuffer;
import java.util.Spliterator;
import java.util.function.DoubleConsumer;
import java.util.stream.DoubleStream;

/**
 * <p>Immutable array of double primitives.</p>
 *
 * <p>Double arrays are constructed by the static interface methods of this interface. The {@code copyOf} method returns
 * a new array representing a shallow deep of the array / buffer argument. The {@code unsafeValueOf}
 * methods wrap a new {@code DoubleArray} around a native array or buffer. It is imperative that this native array /
 * buffer is not modified outside the scope of {@code DoubleArray}, as this would break the immutable contract. The
 * {@code unsafeValueOf} method is used to prevent unnecessary array copies. It is the caller's responsibility not to 
 * violate the contract.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Primitive arrays, such as {@code DoubleArray}, always report {@code NONNULL}. The set of characteristics of 
 * immutable arrays are constant.</p>
 *
 * @author Leon van Zantvoort
 */
@Immutable
public interface DoubleArray extends Array<Double> {

    /** Empty array. */
    static final DoubleArray EMPTY = unsafeValueOf();

    /**
     * Creates an immutable copy of the specified {@code array}, backed by a native array. If the specified
     * {@code array} is mutable, changes to the mutable array are not reflected by this immutable copy.
     *
     * <p>The specified {@code array} can be any array of primitive or object numbers. Reconstructing a copy of this
     * array may involve rounding or as implemented by {@link Number#doubleValue}. The specified {@code array}
     * must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code MUTABLE} and {@code BOUNDED}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return immutable array of double values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    static DoubleArray copyOf(@Nonnull Array<? extends Number> array) {
        double[] backingArray = NativeArrays.doubleArray(array);
        if (backingArray != null) {
            return Arrays.newArray(backingArray, array.offset(), array.length(), array.characteristics(), array.hasCharacteristics(MUTABLE));
        }
        return Arrays.newArray(NativeArrays.toDoubleArray(array), 0, array.length(), array.characteristics(), false);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return immutable array of double values.
     */
    static DoubleArray copyOf(@Nonnull double... array) {
        return copyOf(array, 0, array.length);
    }

    /**
     * <p>Creates an immutable array representation of the specified {@code array}. A defensive copy of {@code array} is
     * used as backing array to ensure immutability.</p>
     *
     * <p>Note that the backing array will have the same length as {@code array}.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return immutable array of double values.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static DoubleArray copyOf(@Nonnull double[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, true);
    }

    /**
     * <p>Returns an immutable array backed by the specified {@code buffer}. All operations on the resulting array are
     * delegated to a duplicate of the buffer. The buffer must be read only.</p>
     *
     * <p>The returned array is serializable. The deserialized array will allocate a direct buffer if the specified
     * buffer is direct.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param buffer the buffer by which the array will be backed.
     * @return an immutable array backed by the specified buffer.
     * @throws NullPointerException if buffer is null.
     */
    public static DoubleArray copyOf(@Nonnull DoubleBuffer buffer) {
        return Arrays.newArray(buffer);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array, it must therefore not be modified after this method returns as this would violate
     * the resulting array's immutability and characteristics contract.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return immutable array that is backed by the specified {@code array}.
     */
    static DoubleArray unsafeValueOf(@Nonnull double... array) {
        return unsafeValueOf(array, 0, array.length);
    }

    /**
     * Creates an immutable array representation of the specified {@code array}. The specified {@code array} is
     * used as backing array, it must therefore not be modified after this method returns as this would violate
     * the resulting array's immutability and characteristics contract.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return immutable array that is backed by the specified {@code array}.
     */
    static DoubleArray unsafeValueOf(@Nonnull double[] array, int offset, int length) {
        return Arrays.newArray(array, offset, length, 0, false);
    }

    /**
     * <p>Returns an immutable array backed by the specified {@code buffer}. All operations on the resulting array are
     * delegated to a duplicate of the buffer.</p>
     *
     * <p>The contents of {@code buffer} must not be modified directly, or indirectly, after this method returns as this
     * would violate this array's immutability and characteristics contract.</p>
     *
     * <p>The returned array is serializable. The deserialized array will allocate a direct buffer if the specified
     * buffer is direct.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param buffer the buffer by which the array will be backed.
     * @return an immutable array backed by the specified buffer.
     * @throws NullPointerException if buffer is null.
     */
    public static DoubleArray unsafeValueOf(@Nonnull DoubleBuffer buffer) {
        return Arrays.newArray(buffer);
    }

    /**
     * Returns the double value found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return double value found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    double getDouble(int index);

    /**
     * Returns {@code true} if this array contains the specified value.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified value.
     */
    boolean containsDouble(double value);

    /**
     * Returns the index of the first occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    int indexOfDouble(double value);

    /**
     * Returns the index of the last occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified value is found.</p>
     *
     * @param value value whose highest index in this array is to be returned.
     * @return highest index of value in array or {@code -1} if value is not found.
     */
    int lastIndexOfDouble(double value);

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     * @throws NullPointerException if consumer is null.
     */
    void forEachDouble(@Nonnull DoubleConsumer consumer);

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    double[] toArray();

    /**
     * Returns an array containing all of the double values of {@code this} array. If the double values fit in the
     * specified array, it is returned therein. Otherwise, a new double array is allocated.
     *
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    double[] toArray(@Nonnull double[] array);

    /**
     * Stores elements from this array into the specified array.
     *
     * @param array the array into which the elements of this array are to be stored.
     * @param offset the offset within the specified array of the first element to be written.
     * @param length the number of elements to be stored. Must be smaller than length of this array.
     * @throws IndexOutOfBoundsException if offset and length parameters are invalid.
     */
    void toArray(@Nonnull double[] array, int offset, int length);

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    DoubleArray offset(int offset);    // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    DoubleArray length(int length);    // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    DoubleArray subArray(int fromIndex, int toIndex);  // More specific return type.

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    ArrayIterator.OfDouble iterator(); // More specific return type.

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    ArrayIterator.OfDouble iterator(int index);   // More specific return type.

    /**
     * Creates a {@code Spliterator} over the elements in this array.
     *
     * <p>The {@code Spliterator} reports:
     * <ul>
     *     <li>{@code Spliterator.SIZED};</li>
     *     <li>{@code Spliterator.SUBSIZED};</li>
     *     <li>{@code Spliterator.ORDERED};</li>
     *     <li>{@code Spliterator.IMMUTABLE} if {@code hasCharacteristics(MUTABLE)} returns {@code false};</li>
     *     <li>{@code Spliterator.NONNULL} if {@code hasCharacteristics(NONNULL)} returns {@code true};</li>
     *     <li>{@code Spliterator.SORTED} if {@code hasCharacteristics(SORTED)} returns {@code true};</li>
     *     <li>{@code Spliterator.DISTINCT} if {@code hasCharacteristics(DISTINCT)} returns {@code true}.</li>
     * </ul>
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     * return Arrays.spliterator(array);
     * </pre>
     *
     * @return a {@code Spliterator} over the elements in this array.
     */
    @Override
    default Spliterator.OfDouble spliterator() {   // More specific return type.
        return Arrays.spliterator(this);
    }

    /**
     * Returns a sequential primitive double stream for this array.
     *
     * @return sequential primitive double stream.
     */
    default DoubleStream doubleStream() {
        return Arrays.doubleStream(this);
    }

    /**
     * Returns a parallel primitive double stream for this array.
     *
     * @return parallel primitive double stream.
     */
    default DoubleStream doubleParallelStream() {
        return Arrays.doubleParallelStream(this);
    }
}
