/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Comparator;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.DoubleConsumer;

import static java.lang.System.arraycopy;
import static java.util.Arrays.copyOfRange;
import static java.util.Objects.requireNonNull;

/**
 * <p>Immutable array of double primitives.</p>
 *
 * <p>This double array is constructed by the static interface methods of {@code DoubleArray}. The {@code copyOf}
 * method returns a new array representing a deep copy of the array argument. The
 * {@code unsafeValueOf} methods wrap a new {@code DoubleArray} around a native array. It is imperative that this 
 * native array is not modified outside the scope of this class, as this would break the immutable contract. 
 * The {@code unsafeValueOf} method is used to prevent unnecessary array copies. It is the caller's responsibility not 
 * to violate the contract.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This 
 * means that this implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. Primitive arrays, such as this class, 
 * always report {@code NONNULL}. The set of characteristics of immutable arrays are constant.</p>
 * 
 * @author Leon van Zantvoort
 */
@Immutable
class DoubleArrayImpl extends AbstractArray<Double> implements NativeArray<Double>, DoubleArray, Serializable {

    /** Backing array. */
    protected transient double[] elements;

    /**
     * Package visible constructor, to be used by static factory methods only.
     *
     * @param array backing array.
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    DoubleArrayImpl(@Nonnull double[] array, int offset, int length, int characteristics, boolean copy) {
        super(copy ? 0 : offset, length, characteristics);
        requireNonNull(array);
        if (length > array.length - offset) {
            throw new IndexOutOfBoundsException();
        }
        if (copy) {
            this.elements = new double[length];
            arraycopy(array, offset, elements, 0, length);
        } else {
            this.elements = array;
        }
    }

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    @Override
    public final Class<Double> getComponentType() {
        return Double.TYPE;
    }

    /**
     * Tells whether or not this array is backed by an accessible array.
     *
     * <p>If this method returns {@code true} then the {@link #nativeArray()} method may safely be invoked.</p>
     *
     * @return {@code true} if, and only if, this array is backed by an accessible array.
     */
    @Override
    public boolean hasArray() {
        return true;
    }

    /**
     * <p>Returns a native array containing all elements of this array.</p>
     *
     * @return native array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public final double[] nativeArray() {
        return elements;
    }

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @Override
    public final @CheckForNull Comparator<? super Double> getComparator() {
        return null;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return characteristics | NONNULL & ~(MUTABLE | BOUNDED | UNBOUNDED);
    }

    /**
     * Returns the object found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public Double get(int index) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        return getDouble(index);
    }

    /**
     * Returns the double value found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return double value found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public double getDouble(int index) {
        rangeCheck(index);
        return elements[offset + index];
    }

    /**
     * Returns {@code true} if this array contains the specified object. This method supports boxed primitives. This
     * means that if this array represents a primitive array, this method returns {@code true} if this array contains
     * the unboxed version of specified object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object object whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified object.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean contains(Object object) {
        if (object instanceof Double) {
            if (Tripwire.ENABLED) {
                Tripwire.trip(getClass());
            }
            return containsDouble((Double) object);
        }
        return false;
    }

    /**
     * Returns {@code true} if this array contains the specified value.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified value.
     */
    @Override
    public boolean containsDouble(double value) {
        if (hasCharacteristics(SORTED)) {
            return Arrays.binarySearch(this, value) >= 0;
        }
        final double[] a = elements;
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            if (equals(a[index], value)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns {@code true} if this array contains all of the elements in the specified {@code array}. Primitives and
     * their object counterparts are regarded equal in this respect.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array is sorted, otherwise all elements are
     * traversed until the specified value is found.</p>
     *
     * @param array array to be checked for containment in this array.
     * @return {@code true} if this array contains all of the elements in the specified array.
     * @throws NullPointerException if array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean containsAll(@Nonnull Array<?> array) {
        // Fastest: direct array access.
        double[] backingArray = NativeArrays.doubleArray(array);
        if (backingArray != null) {
            for (int index = array.offset(), toIndex = index + array.length(); index < toIndex; index++) {
                if (!containsDouble(backingArray[index])) {
                    return false;
                }
            }
            return true;
        }
        // Fast alternative: no unboxing.
        DoubleArray cast = ArrayCast.toDoubleArray(array);
        if (cast != null) {
            final int length = array.length();
            final ArrayIterator.OfDouble iterator = cast.iterator(length);
            for (int index = 0; index < length; index++) {
                if (!containsDouble(iterator.nextDouble())) {
                    return false;
                }
            }
            return true;
        }
        // Alternative: unboxing.
        for (Object o : array) {
            if (!(o instanceof Double) || !containsDouble((Double) o)) {
                return false;   // Returns false for null elements.
            }
        }
        return true;
    }

    /**
     * Returns the index of the first occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public int indexOf(Object object) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        if (object instanceof Double) {
            return indexOfDouble((Double) object);
        }
        return -1;
    }

    /**
     * Returns the index of the first occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    @Override
    public int indexOfDouble(double value) {
        if (hasCharacteristics(SORTED)) {
            int index = Arrays.binarySearch(this, value);
            if (index >= 0) {
                return Arrays.binarySearchFirst(this, index);
            }
            return -1;
        }
        final double[] a = elements;
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            if (equals(a[index], value)) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified object is found.</p>
     *
     * @param object object whose highest index in this array is to be returned.
     * @return highest index of object in array or {@code -1} if value is not found.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public int lastIndexOf(Object object) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        if (object instanceof Double) {
            return lastIndexOfDouble((Double) object);
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified value is found.</p>
     *
     * @param value value whose highest index in this array is to be returned.
     * @return highest index of value in array or {@code -1} if value is not found.
     */
    @Override
    public int lastIndexOfDouble(double value) {
        if (hasCharacteristics(SORTED)) {
            int index = Arrays.binarySearch(this, value);
            if (index >= 0) {
                return Arrays.binarySearchLast(this, index);
            }
            return -1;
        }
        final double[] a = elements;
        for (int toIndex = offset + length, index = toIndex - 1; index >= offset; index--) {
            if (equals(a[index], value)) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     * @throws NullPointerException if consumer is null.
     */
    @Override
    public void forEach(@Nonnull Consumer<? super Double> consumer) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        forEachDouble(consumer::accept);
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     * @throws NullPointerException if consumer is null.
     */
    @Override
    public void forEachDouble(@Nonnull DoubleConsumer consumer) {
        requireNonNull(consumer);
        final double[] a = elements;
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            consumer.accept(a[index]);
        }
    }

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    public double[] toArray() {
        return copyOfRange(elements, offset, offset + length);
    }

    /**
     * Returns an array containing all of the double values of {@code this} array. If the double values fit in the
     * specified array, it is returned therein. Otherwise, a new double array is allocated.
     *
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    @Override
    public double[] toArray(@Nonnull double[] array) {
        double[] a = array.length >= length ? array : new double[length];
        arraycopy(elements, offset, a, 0, length);
        return a;
    }

    /**
     * Stores elements from this array into the specified array.
     *
     * @param array the array into which the elements of this array are to be stored.
     * @param offset the offset within the specified array of the first element to be written.
     * @param length the number of elements to be stored. Must be smaller than length of this array.
     * @throws IndexOutOfBoundsException if offset and length parameters are invalid.
     */
    @Override
    public void toArray(@Nonnull double[] array, int offset, int length) {
        arraycopy(elements, this.offset, array, offset, length);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public DoubleArray offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new DoubleArrayImpl(elements, newOffset, newLength, characteristics, false);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public DoubleArray length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new DoubleArrayImpl(elements, offset, length, characteristics, false);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public DoubleArray subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new DoubleArrayImpl(elements, newOffset, newLength, characteristics, false);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator.OfDouble {
        int index;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @Override
        public double nextDouble() {
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i + 1;
            return elements[i];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @Override
        public double previousDouble() {
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i;
            return elements[i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @Override
        public void setDouble(double value) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void addDouble(double value) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator.OfDouble iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator.OfDouble iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(index);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public String toString() {
        int iMax = offset + length - 1;
        if (iMax - offset == -1) {
            return "[]";
        }
        StringBuilder b = new StringBuilder();
        b.append('[');
        for (int i = offset; ; i++) {
            b.append(elements[i]);
            if (i == iMax) {
                return b.append(']').toString();
            }
            b.append(", ");
        }
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public int hashCode() {
        int result = 1;
        int to = offset + length;
        for (int i = offset; i < to; i++) {
            long bits = Double.doubleToLongBits(elements[i]);
            result = 31 * result + (int) (bits ^ (bits >>> 32));
        }
        return result;
   }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj instanceof Array) {
            Array<?> array = (Array<?>) obj;
            if (length != array.length()) {
                return false;
            }
            // Fastest: direct array access.
            double[] backingArray = NativeArrays.doubleArray(array);
            if (backingArray != null) {
                final int o = array.offset();
                for (int i = 0; i < length; i++) {
                    if (!equals(elements[offset + i], backingArray[o + i])) {
                        return false;
                    }
                }
                return true;
            }
            // Fast alternative: no unboxing.
            DoubleArray cast = ArrayCast.toDoubleArray(array);
            if (cast != null) {
                final ArrayIterator.OfDouble iterator = cast.iterator();
                for (int index = offset, toIndex = index + length; index < toIndex; index++) {
                    if (!equals(elements[index], iterator.nextDouble())) {
                        return false;
                    }
                }
                return true;
            }
            // Alternative.
            return Arrays.equals(this, array);
        }
        return false;
    }

    /**
     * Returns {@code true} if both values are equal, {@code false} otherwise.
     *
     * @param x first value to check.
     * @param y second value to check.
     * @return {@code true} if both values are equal, {@code false} otherwise.
     */
    private static boolean equals(double x, double y) {
        return Arrays.equals(x, y);
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    Object writeReplace() {
        return new SerializationProxy(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy implements Serializable {
        static final double serialVersionUID = -534852546189583117L;

        transient int offset;
        final int length;
        final int characteristics;
        transient double[] elements;

        SerializationProxy(DoubleArrayImpl array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int index = offset, toIndex = index + length; index < toIndex; index++) {
                s.writeDouble(elements[index]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = new double[length];
            for (int index = 0; index < length; index++) {
                elements[index] = s.readDouble();
            }
        }

        private Object readResolve() {
            return new DoubleArrayImpl(elements, 0, length, characteristics, false);
        }
    }
}
