/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import java.util.Comparator;
import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;
import java.util.function.Consumer;

import static java.lang.System.arraycopy;
import static java.util.Objects.requireNonNull;

/**
 * <p>Dynamic-size array of object references.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Dynamic arrays, such as this class, always report {@code MUTABLE}.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that the array implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. Characteristics are always invalidated if
 * conditions of these characteristics are broken as result of array modification. Characteristics are only added in
 * scenarios in which their conditions can be 'tested' without traversal of elements. For example, the {@code SORTED}
 * characteristic is added if the length of an array is reduced to zero or one element, or if all elements are replaced
 * by an array that is marked as being sorted. The semantics of characteristics processing is documented for each
 * "destructive" method. Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * <p>This class eliminates obsolete references by default to allow the garbage collector to reclaim memory. This is
 * done by setting the elements in 'reserved space' to {@code null}. This feature can be disabled by calling
 * {@code setClearReservedElements(false)}.</p>
 *
 * @author Leon van Zantvoort
 */
@NotThreadSafe
abstract class DynamicObjectArrayImpl<E> extends MutableObjectArrayImpl<E> implements DynamicObjectArray<E> {

    /** Specifies if reserved elements need to be reset to default values. */
    protected boolean clearReservedElements = false;

    /** Counter for fail-fast error detection. */
    protected transient int modCount = 0;

    /**
     * Package visible constructor used to wrap specified {@code array}.
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param characteristics array characteristics.
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> is used.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    DynamicObjectArrayImpl(@Nonnull E[] array, int offset, int length, int characteristics,
                           @Nullable Comparator<? super E> comparator, boolean copy) {
        super(array, offset, length, characteristics, comparator, copy);
    }

    /**
     * Specifies if reserved elements need to be reset to default values. This flag is set to {@code false} by default
     * for primitive arrays.
     *
     * @param flag {@code true} if reserved elements need to be reset, {@code false} otherwise.
     */
    @Override
    public void setClearReservedElements(boolean flag) {
        // Must not be invoked as sub array.
        clearReservedElements = flag;
    }

    /**
     * Returns {@code true} if reserved elements need to be reset to default values, {@code false} otherwise.
     *
     * <p>Sub arrays must override this method and delegate the invocation to the parent.</p>
     *
     * @return {@code true} if reserved elements need to be reset, {@code false} otherwise.
     */
    protected boolean clearReservedElements() {
        return clearReservedElements;
    }

    /**
     * Fail-fast error detection. Sub classes (such as sub arrays) may override this method if they wish to implement
     * concurrent modification checking.
     *
     * @throws ConcurrentModificationException if parent array has been structurally modified.
     */
    protected void checkForComodification() {
    }

    /**
     * Updates length of this array and increases modCount.
     *
     * <p>Sub arrays must override this method and delegate the invocation this implementation by calling
     * {@code super.updateLength()}, next they must delegate the invocation to the parent array.</p>
     *
     * @param diff number of elements that are added / removed.
     */
    protected void updateLength(int diff) {
        modCount++;
        if (clearReservedElements && diff < 0) {    // Following if statement is always false for sub arrays.
            int newLength = length + diff;
            while (length != newLength) {
                elements[offset + --length] = null;
            }
        } else {
            length += diff;
        }
    }

    /**
     * Returns the number of additional elements that this array can store (in the absence of memory or resource
     * constraints), or {@code Integer.MAX_VALUE} if there is no intrinsic limit.
     *
     * @return the remaining capacity.
     */
    @Override
    public int remainingCapacity() {
        return elements.length - offset - length;
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     * @throws NullPointerException if consumer is null.
     */
    @SuppressWarnings({"unchecked", "ConstantConditions"})
    @Override
    public void forEach(@Nonnull Consumer<? super E> consumer) {
        requireNonNull(consumer);
        final int expectedModCount = modCount;
        final Object[] a = elements;
        for (int index = offset, toIndex = index + length; modCount == expectedModCount && index < toIndex; index++) {
            consumer.accept((E) a[index]);
        }
        if (modCount != expectedModCount) {
            throw new ConcurrentModificationException();
        }
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. The semantics of this method are less
     * strict than the semantics expressed by {@code MutableArray}; The length of the array that is passed as argument
     * needs to be less than or equal to this array's capacity.
     *
     * <p>The length of {@code this} array is increased or decreased by length difference between {@code this} array and
     * {@code array} argument.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws NullPointerException if the specified array is null.
     * @throws IllegalStateException if elements cannot be replaced due to capacity issues.
     */
    @Override
    public void setAll(@Nonnull Array<? extends E> array) {
        int arrayLength = array.length();
        if (arrayLength <= elements.length - offset) {
            int lengthDiff = arrayLength - length;
            ObjectArray<?> cast = ArrayCast.toObjectArray(array);
            if (cast != null) {
                cast.toArray(elements, offset, arrayLength);
            } else {
                arraycopy(NativeArrays.toArray(array), 0, elements, offset, arrayLength);
            }
            updateLength(lengthDiff);
            characteristics(array.characteristics());   // All elements are replaced by new array; Copy characteristics.
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(offset, array);
            }
        } else {
            throw new IllegalStateException("Array full.");
        }
    }

    /**
     * Appends the specified object to the end of this array. Length of this array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param object object to to be added.
     * @throws IllegalStateException if element cannot be inserted due to capacity issues.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void add(@Nullable E object) {
        int elementIndex = offset + length;
        if (elementIndex < elements.length) {
            elements[elementIndex] = object;
            updateLength(1);
            if (object == null) {
                removeCharacteristics(NONNULL);
            }
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(elementIndex, object);
            }
        } else {
            throw new IllegalStateException("Array full.");
        }
    }

    /**
     * Appends the elements of the specified array to the end of this array. Length of {@code this} array is increased
     * by {@code array.length()}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED} and if it does not break the sorted order of
     *   this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is reported
     *   after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT} and if the
     *   specified {@code array} does not contain any objects that were already present in this array.</li>
     * </ul>
     *
     * @param array array to be added.
     * @throws NullPointerException if the specified array is null.
     * @throws IllegalStateException if elements cannot be inserted due to capacity issues.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void addAll(@Nonnull Array<? extends E> array) {
        int arrayLength = array.length();
        int elementIndex = offset + length;
        if (arrayLength > 0) {
            if (arrayLength <= elements.length - elementIndex) {
                ObjectArray<?> cast = ArrayCast.toObjectArray(array);
                if (cast != null) {
                    cast.toArray(elements, elementIndex, arrayLength);
                } else {
                    arraycopy(NativeArrays.toArray(array), 0, elements, elementIndex, arrayLength);
                }
                updateLength(arrayLength);
                retainCharacteristics(array.characteristics());
                if (hasCharacteristics(SORTED)) {
                    updateCharacteristics(elementIndex, array);
                }
            } else {
                throw new IllegalStateException("Array full.");
            }
        }
    }

    /**
     * Removes last element. Length of this array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains no elements after this invocation, {@code NONNULL}, {@code SORTED} and
     *   {@code DISTINCT} are reported.
     * </ul>
     *
     * @return element that is removed.
     * @throws NoSuchElementException if array is already empty.
     */
    @SuppressWarnings("unchecked")
    @Override
    public E remove() {
        if (length == 0) {
            throw new NoSuchElementException();
        }
        int elementIndex = offset + length - 1;
        E oldValue = (E) elements[elementIndex];
        updateLength(-1);
        return oldValue;
    }

    /**
     * Removes all elements.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>After this invocation, {@code NONNULL}, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     */
    @Override
    public void clear() {
        if (length > 0) {
            updateLength(-length);
        }
        addCharacteristics(NONNULL | SORTED | DISTINCT);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        int index;
        int lastIndex = -1;

        /**
         * The modCount value that the iterator believes that the backing
         * array should have. If this expectation is violated, the iterator
         * has detected concurrent modification.
         */
        int expectedModCount = modCount;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        void checkForComodification() {
            if (modCount != expectedModCount) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @SuppressWarnings("unchecked")
        @Override
        public E next() {
            checkForComodification();
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            E[] a = (E[]) elements;
            if (i >= a.length) {
                throw new ConcurrentModificationException();
            }
            index = i + 1;
            return a[lastIndex = i];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @SuppressWarnings("unchecked")
        @Override
        public E previous() {
            checkForComodification();
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            E[] a = (E[]) elements;
            if (i >= a.length) {
                throw new ConcurrentModificationException();
            }
            index = i;
            return a[lastIndex = i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @SuppressWarnings("unchecked")
        @Override
        public void set(E object) {
            if (lastIndex - offset < 0) {
                throw new IllegalStateException();
            }
            checkForComodification();
            try {
                DynamicObjectArrayImpl.this.set(lastIndex - offset, object);
            } catch (IndexOutOfBoundsException e) {
                throw new ConcurrentModificationException();
            }
        }

        @Override
        public void add(E object) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(index);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public abstract DynamicObjectArray<E> offset(int offset);
}
