/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

/**
 * Helper interface for modifying array characteristics.
 *
 * @author Leon van Zantvoort
 */
interface MutableArrayCharacteristics {

    // This interfaces does not extend ArrayCharacteristics to prevent it from being reported as sub interface of
    // ArrayCharacteristics in JavaDoc.

    /**
     * Returns the characteristics set for this array. The result is represented as ORed values from
     * {@code NONNULL}, {@code MUTABLE}, {@code BOUNDED}, {@code SORTED}, {@code DISTINCT}.
     * Modification of the array contents may result in updated characteristics.
     *
     * @return the characteristics set for this array.
     */
    int characteristics();

    /**
     * Sets the specified characteristics for this array.
     *
     * @param characteristics characteristics to be set for this array.
     * @return the characteristics that were actually stored for this array.
     */
    int characteristics(int characteristics);

    /**
     * Retains the characteristics for this array that are present in the specified parameter. Characteristics that
     * are not part of the array are removed.
     *
     * @param characteristics the characteristics that need to be retained for this array.
     * @return the characteristics that were actually stored for this array.
     */
    default int retainCharacteristics(int characteristics) {
        if (characteristics == ~0) { // Optimization.
            return characteristics();
        }
        return characteristics(characteristics() & characteristics);
    }

    /**
     * Adds the specified characteristics for this array.
     *
     * @param characteristics the characteristics that need to be added for this array.
     * @return the characteristics that were actually stored for this array.
     */
    default int addCharacteristics(int characteristics) {
        if (characteristics == 0) { // Optimization.
            return characteristics();
        }
        return characteristics(characteristics() | characteristics);
    }

    /**
     * Removes the specified characteristics for this array.
     *
     * @param characteristics the characteristics that need to be removed for this array.
     * @return the characteristics that were actually stored for this array.
     */
    default int removeCharacteristics(int characteristics) {
        if (characteristics == 0) { // Optimization.
            return characteristics();
        }
        return characteristics(characteristics() & ~characteristics);
    }
}
