/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import org.squarebrackets.function.BooleanUnaryOperator;

import javax.annotation.Nonnull;

import static java.util.Objects.requireNonNull;

/**
 * <p>Fixed-size mutable array of boolean primitives.</p>
 *
 * <p>Mutable boolean arrays are constructed by the static interface methods of this interface. The {@code copyOf}
 * method returns a new array representing a deep copy of the array argument. The {@code valueOf} methods wrap a new
 * {@code MutableBooleanArray} around a native array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Mutable primitive arrays, such as {@code MutableBooleanArray}, always report {@code MUTABLE} and {@code NONNULL}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 *
 * @author Leon van Zantvoort
 */
public interface MutableBooleanArray extends BooleanArray, MutableArray<Boolean> {

    /**
     * Creates a fixed-size mutable copy of the specified {@code array}, backed by a native array. If the
     * specified {@code array} is mutable, changes to the mutable array are not reflected by this mutable copy.
     *
     * <p>The specified {@code array} can be an array of {@code boolean} primitives or {@code Boolean} objects. The
     * specified {@code array} must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL} and {@code MUTABLE};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code BOUNDED}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return fixed-size mutable array of boolean values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static MutableBooleanArray copyOf(@Nonnull Array<Boolean> array) {
        boolean[] backingArray = NativeArrays.booleanArray(array);
        if (backingArray != null) {
            return Arrays.newMutableArray(backingArray, array.offset(), array.length(), array.characteristics(), true);
        }
        return Arrays.newMutableArray(NativeArrays.toBooleanArray(array), 0, array.length(), array.characteristics(), false);
    }

    /**
     * Creates a mutable array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p>The contents of {@code array} must not be modified after this method returns as this would violate this
     * array's characteristics contract.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL} and {@code MUTABLE};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return a mutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    static MutableBooleanArray valueOf(@Nonnull boolean... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates a mutable array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p>The contents of {@code array} must not be modified after this method returns as this would violate this
     * array's characteristics contract.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL} and {@code MUTABLE};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return a mutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static MutableBooleanArray valueOf(@Nonnull boolean[] array, int offset, int length) {
        return Arrays.newMutableArray(array, offset, length, 0, false);
    }

    /**
     * Replaces element at specified {@code index} by specified boolean value.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index index relative to the offset of the array.
     * @param value boolean value to replace existing value.
     * @return boolean value that is replaced.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    boolean setBoolean(int index, boolean value);

    /**
     * Replaces each element of this list with the result of applying the operator to that element. Errors or runtime
     * exceptions thrown by the operator are relayed to the caller.
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     *{@literal for (int i = 0; i < array.length(); i++) {
     *     setBoolean(i, operator.applyAsBoolean(a.getBoolean(i));
     * }
     * }</pre>
     *
     * @param operator the operator to apply to each element.
     * @throws NullPointerException if the specified operator is null.
     */
    default void replaceAllBoolean(@Nonnull BooleanUnaryOperator operator) {
        requireNonNull(operator);
        int toIndex = length();
        for (int i = 0; i < toIndex; i++) {
            setBoolean(i, operator.applyAsBoolean(getBoolean(i)));
        }
    }
    
    /**
     * Returns the backing array (optional operation). All elements of {@code this} mutable array and parents of this
     * mutable array are stored in the returned object. Modifications to the backing array are reflected by {@code this}
     * mutable array and it's parents.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing class.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>This method allows the caller to modify this array outside the scope of the implementing class. Hence, it can
     * not guarantee correctness of the characteristics of this array. For that reason, content related characteristics
     * (such as {@code SORTED} and {@code DISTINCT}) are removed by calling this method.
     * Characteristics removal can be suppressed by obtaining the backing array through {@code doAction}:</p>
     *
     * <pre>
     *{@literal boolean[] backingArray = (boolean[]) array.doAction(ArrayContext::array);}</pre>
     *
     * @return backing array.
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    boolean[] array();  // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    MutableBooleanArray offset(int offset);     // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    MutableBooleanArray length(int length);     // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    MutableBooleanArray subArray(int fromIndex, int toIndex);   // More specific return type.
}
