/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.nio.Buffer;
import java.util.Comparator;
import java.util.List;
import java.util.ListIterator;
import java.util.Objects;

/**
 * <p>Mutable, fixed-size, wrapper for Java Collection Framework's List.</p>
 *
 * <p>This mutable object array is constructed by the static {@code newMutableArray} methods of {@code Arrays}.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Mutable arrays, such as this class, always report {@code MUTABLE}.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that this implementation determines which set of characteristics apply to its elements without traversing these
 * elements, in other words, without incurring a performance penalty. Characteristics are always invalidated if
 * conditions of these characteristics are broken as result of array modification. Characteristics are only added in
 * scenarios in which their conditions can be 'tested' without traversal of elements. For example, the {@code SORTED}
 * characteristic is added if the length of an array is reduced to zero or one element, or if all elements are replaced
 * by an array that is marked as being sorted. The semantics of characteristics processing is documented for each
 * "destructive" method. Users can take full control of characteristics assignment by means of the
 * {@link MutableArray#doAction(ArrayAction)} method.</p>
 *
 * @author Leon van Zantvoort
 */
@NotThreadSafe
class MutableListArrayImpl<E> extends ListArrayImpl<E> implements MutableObjectArray<E>, MutableArrayCharacteristics {

    /** Is {@code true} if control is transferred to caller of {@code doAction}. */
    protected transient boolean inAction;

    /** Allows caller of {@code doAction} to control state of array. */
    private transient ArrayContext<E> context;

    /**
     * Package visible constructor.
     *
     * @param list backing list.
     * @param offset start index of range.
     * @param length length of range.
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> is used.
     * @param characteristics array characteristics.
     */
    MutableListArrayImpl(@Nonnull List<E> list, int offset, int length, int characteristics,
                         @Nullable Comparator<? super E> comparator) {
        super(list, offset, length, characteristics, comparator);
    }

    /**
     * Returns the array context of this array, or {@code null} if control is already transferred to caller of
     * {@code doAction}.
     *
     * @return array context.
     */
    protected ArrayContext<E> getArrayContext() {
        if (inAction) {
            return null;
        }
        if (context == null) {
            context = new ArrayContextImpl();   // Lazy initialization.
        }
        return context;
    }

    /**
     * Updates the SORTED and DISTINCT characteristics of this array by looking at the elements surrounding the
     * specified {@code value}. The SORTED and/or DISTINCT characteristics are removed if the value invalidates the
     * characteristics. This method must only be invoked when this array reports the {@code SORTED} characteristic.
     *
     * @param elementIndex index relative to the backing array.
     * @param object object that is stored at index.
     */
    @SuppressWarnings("unchecked")
    protected void updateCharacteristics(int elementIndex, E object) {
        assert hasCharacteristics(SORTED);
        if (length > 1) {
            Comparator<Object> c = null;
            if (comparator == null) {
                if (object != null && object instanceof Comparable) {
                    c = Arrays.NATURAL_ORDER_COMPARATOR;
                } else {
                    removeCharacteristics(SORTED | DISTINCT);
                }
            } else {
                c = (Comparator<Object>) comparator;
            }
            if (c != null) {
                if (elementIndex > offset) {
                    if (c.compare(object, elements.get(elementIndex - 1)) < 0) {
                        removeCharacteristics(SORTED | DISTINCT);
                    } else if (c.compare(object, elements.get(elementIndex - 1)) == 0) {
                        removeCharacteristics(DISTINCT);
                    }
                }
                if (elementIndex < offset + length - 1) {
                    if (c.compare(object, elements.get(elementIndex + 1)) > 0) {
                        removeCharacteristics(SORTED | DISTINCT);
                    } else if (c.compare(object, elements.get(elementIndex + 1)) == 0) {
                        removeCharacteristics(DISTINCT);
                    }
                }
            }
        }
    }

    /**
     * Updates the SORTED and DISTINCT characteristics of this array by looking at the elements surrounding the
     * first and last element of the specified {@code array}. The SORTED and/or DISTINCT characteristics are removed if
     * these elements invalidate the characteristics.
     *
     * @param elementIndex index relative to the backing array.
     * @param array array that is stored, starting at index.
     */
    protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends E> array) {
        assert hasCharacteristics(SORTED);
        int len = array.length();
        if (length > len) {
            if (len > 0) {
                updateCharacteristics(elementIndex, elements.get(elementIndex));
            }
            if (len > 1) {
                if (hasCharacteristics(SORTED)) {
                    updateCharacteristics(elementIndex + len - 1, elements.get(elementIndex + len - 1));
                }
            }
        }
    }

    /**
     * Sets the specified characteristics for this array.
     *
     * @param characteristics characteristics to be set for this array.
     * @return the characteristics that were actually stored for this array.
     */
    @Override
    public int characteristics(int characteristics) {
        if (characteristics != this.characteristics) { // Optimization.
            if (!inAction) {
                this.characteristics = filterCharacteristics(characteristics);
            }
        }
        return this.characteristics;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    @Override
    protected int filterCharacteristics(int characteristics) {
        return characteristics | MUTABLE & ~(BOUNDED | UNBOUNDED);
    }

    /**
     * Performs the specified {@code action} without affecting the array's set of characteristics and comparator.
     * Ownership of array characteristics / comparator control is moved to the caller by means of the
     * {@code ArrayContext} object. Errors or runtime exceptions thrown by the action are relayed to the caller.
     *
     * @param <T> return type.
     * @param action the action to be performed.
     * @return the value returned by the action's {@code run} method.
     * @throws IllegalStateException if action is called when array does not own control of characteristics / comparator.
     * @exception NullPointerException if the action is {@code null}.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public final <T> T doAction(@Nonnull ArrayAction<E, T> action) {
        ArrayContext<E> context = getArrayContext();
        if (context == null) {
            throw new IllegalStateException("Already in action.");
        }
        try {
            inAction = true;
            return action.run(context);
        } finally {
            inAction = false;
        }
    }

    /**
     * Tells whether or not this array is backed by an accessible array.
     *
     * <p>If this method returns {@code true} then the {@link #array()} method may safely be invoked.</p>
     *
     * @return {@code true} if, and only if, this array is backed by an accessible array.
     */
    @Override
    public boolean hasArray() {
        return false;
    }

    /**
     * Throws an {@code UnsupportedOperationException} as this implementation does not expose backing array.
     *
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public Object[] array() {
        throw new UnsupportedOperationException("Backing array not exposed by implementation.");
    }

    /**
     * Tells whether or not this array is backed by an accessible buffer.
     *
     * <p>If this method returns {@code true} then the {@link #buffer()} method may safely be invoked.</p>
     *
     * @return {@code true} if, and only if, this array is backed by an accessible buffer.
     */
    @Override
    public boolean hasBuffer() {
        return false;
    }

    /**
     * Throws an {@code UnsupportedOperationException} as this implementation does not expose backing buffer.
     *
     * @throws UnsupportedOperationException if array does not expose backing buffer.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public Buffer buffer() {
        throw new UnsupportedOperationException("Array not backed by buffer.");
    }

    /**
     * Sets the comparator that is used for sorting this array.
     *
     * @param comparator that is used for sorting this array.
     */
    protected void setComparator(@Nullable Comparator<? super E> comparator) {
        if (comparator != this.comparator) { // Optimization.
            ArrayContext<E> context = getArrayContext();
            if (context != null) {
                context.setComparator(comparator);
            }
        }
    }

    /**
     * Replaces element at specified {@code index} by specified object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code object} is not {@code null};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array. Arrays
     *   sorted according to natural order are not regarded sorted if they contain {@code null} element(s);</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index index relative to the offset of the array.
     * @param object object to replace existing value.
     * @return object that is replaced.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if specified {@code object} is null and array does not permit null values.
     */
    @Override
    public @CheckForNull E set(int index, E object) {
        rangeCheck(index);
        int elementIndex = offset + index;
        E oldObject = elements.set(elementIndex, object);
        if (object == null) {
            removeCharacteristics(NONNULL);
        }
        if (hasCharacteristics(SORTED)) {
            updateCharacteristics(elementIndex, object);
        }
        return oldObject;
    }

    /**
     * Replaces all elements of this array by the elements of the specified array. Note that the length of {@code array}
     * needs to be equal to the length of this array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code NONNULL} prior to this call, {@code NONNULL} is reported after this
     *   invocation if the specified {@code array} reports {@code NONNULL};</li>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code array} reports {@code SORTED};</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code array} reports {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param array array to replace elements of this array.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     * @throws IllegalStateException if length of the specified array is not equal to length of this array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public void setAll(@Nonnull Array<? extends E> array) {
        int lengthDiff = array.length() - length;
        if (lengthDiff == 0) {
            elements.subList(offset, offset + length).clear();
            elements.addAll(offset, ArrayBridge.asList(array));
            characteristics(array.characteristics());   // All elements are replaced by new array; Copy characteristics.
            if (hasCharacteristics(SORTED)) {
                updateCharacteristics(offset, array);
            }
        } else {
            throw new IllegalStateException("Array length.");
        }
    }

    /**
     * Sorts this array according to natural ordering.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @throws NullPointerException if this array contains null elements.
     * @throws ClassCastException if this array contains elements that are not mutually {@code Comparable}.
     */
    @Override
    public void sort() {
        if (!hasCharacteristics(SORTED) || !Objects.equals(comparator, null)) {
            elements.subList(offset, offset + length).sort(Arrays.NATURAL_ORDER_COMPARATOR);
            setComparator(null);
            addCharacteristics(SORTED);  // Array is guaranteed to be sorted.
        }
    }

    /**
     * Sorts this array according to natural ordering. The underlying implementation utilizes multiple threads
     * to reduce sorting time. The algorithm may require a working space equal to the size of this array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @throws NullPointerException if this array contains null elements.
     * @throws ClassCastException if this array contains elements that are not mutually {@code Comparable}.
     */
    @SuppressWarnings("unchecked")
    @Override
    public void parallelSort() {
        if (!hasCharacteristics(SORTED) || !Objects.equals(comparator, null)) {
            Object[] a = toArray();
            java.util.Arrays.parallelSort(a, Arrays.NATURAL_ORDER_COMPARATOR);
            ListIterator it = elements.listIterator(offset);
            for (int index = 0; index < length; index++) {
                it.next();
                it.set(a[index]);
            }
            setComparator(null);
            addCharacteristics(SORTED);  // Array is guaranteed to be sorted.
        }
    }

    /**
     * Sorts this array using the supplied {@code comparator} to compare elements. The {@code comparator} can be
     * retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> should be used.
     * @throws ClassCastException if the list contains elements that are not <i>mutually comparable</i> using the
     * specified comparator.
     * @throws IllegalArgumentException (optional) if the comparator is found to violate the
     * {@link java.util.Comparator} contract.
     * @throws NullPointerException if the specified comparator is null.
     */
    @Override
    public void sort(Comparator<? super E> comparator) {
        if (!hasCharacteristics(SORTED) || !Objects.equals(this.comparator, comparator)) {
            elements.subList(offset, offset + length).sort(comparator);
            setComparator(comparator);
            addCharacteristics(SORTED);  // Array is guaranteed to be sorted.
        }
    }

    /**
     * Sorts this array using the supplied {@code Comparator} to compare elements. The underlying implementation
     * utilizes multiple threads to reduce sorting time. The algorithm requires a working space equal to the size of the
     * this array. The {@code comparator} can be retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> should be used.
     * @throws ClassCastException if the list contains elements that are not <i>mutually comparable</i> using the
     * specified comparator.
     * @throws IllegalArgumentException (optional) if the comparator is found to violate the
     * {@link java.util.Comparator} contract.
     * @throws NullPointerException if the specified comparator is null.
     */
    @SuppressWarnings("unchecked")
    @Override
    public void parallelSort(Comparator<? super E> comparator) {
        if (!hasCharacteristics(SORTED) || !Objects.equals(this.comparator, comparator)) {
            Object[] a = toArray();
            java.util.Arrays.parallelSort(a, (Comparator) comparator);
            ListIterator it = elements.listIterator(offset);
            for (int index = 0; index < length; index++) {
                it.next();
                it.set(a[index]);
            }
            setComparator(comparator);
            addCharacteristics(SORTED);  // Array is guaranteed to be sorted.
        }
    }

    /**
     * Array context implementation.
     */
    @SuppressWarnings("MissingMethodJavaDoc")
    private class ArrayContextImpl implements ArrayContext<E> {

        @Override
        public boolean hasArray() {
            return MutableListArrayImpl.this.hasArray();
        }

        @Override
        public Object array() {
            return MutableListArrayImpl.this.array();
        }

        @Override
        public boolean hasBuffer() {
            return MutableListArrayImpl.this.hasBuffer();
        }

        @Override
        public Buffer buffer() {
            return MutableListArrayImpl.this.buffer();
        }

        @Override
        public int characteristics(int newCharacteristics) {
            return characteristics = filterCharacteristics(newCharacteristics);
        }

        @Override
        public int characteristics() {
            return characteristics;
        }

        @Override
        public Comparator<? super E> getComparator() {
            return comparator;
        }

        @Override
        public void setComparator(@Nullable Comparator<? super E> newComparator) {
            comparator = newComparator;
        }
    }

    /**
     * Sub arrays share serialization proxy with parent arrays intentionally. As a result, deserialized sub arrays become
     * detached from their originating parent.
     */
    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SubMutableListArrayImpl<E> extends MutableListArrayImpl<E> {

        final MutableListArrayImpl<E> parent;

        @SuppressWarnings("unchecked")
        SubMutableListArrayImpl(@Nonnull MutableListArrayImpl<E> parent, int offset, int length) {
            super(parent.elements, offset, length, parent.characteristics, parent.comparator);
            this.parent = parent;
        }

        @Override
        protected ArrayContext<E> getArrayContext() {
            return parent.inAction ? null : super.getArrayContext();
        }

        @Override
        protected void updateCharacteristics(int elementIndex, E object) {
            super.updateCharacteristics(elementIndex, object);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, object);
            }
        }

        @Override
        protected void updateCharacteristics(int elementIndex, @Nonnull Array<? extends E> array) {
            super.updateCharacteristics(elementIndex, array);
            if (hasCharacteristics(SORTED)) {
                parent.updateCharacteristics(elementIndex, array);
            }
        }

        @Override
        protected void setComparator(@Nullable Comparator<? super E> comparator) {
            // Comparator change of sub array means that sorted status of parent array must be reset.
            super.setComparator(comparator);
            parent.removeCharacteristics(SORTED);
            parent.setComparator(null);
        }

        @Override
        public int characteristics(int characteristics) {
            // Parent array retains characteristics that are shared with sub array. Additional characteristics
            // assigned to sub array are not set for parent array, as it is not guaranteed that these
            // characteristics hold for parent array.
            parent.retainCharacteristics(characteristics);
            return super.characteristics(characteristics);
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public MutableObjectArray<E> offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new SubMutableListArrayImpl<>(this, newOffset, newLength);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public MutableObjectArray<E> length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new SubMutableListArrayImpl<>(this, offset, length);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public MutableObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new SubMutableListArrayImpl<>(this, newOffset, newLength);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        final ListIterator<E> iterator;

        ArrayIteratorImpl(ListIterator<E> iterator) {
            this.iterator = iterator;
        }

        @Override
        public boolean hasNext() {
            return iterator.hasNext();
        }

        @Override
        public E next() {
            return iterator.next();
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return iterator.hasPrevious();
        }

        @Override
        public E previous() {
            return iterator.previous();
        }

        @Override
        public int nextIndex() {
            return iterator.nextIndex();
        }

        @Override
        public int previousIndex() {
            return iterator.previousIndex();
        }

        @Override
        public void set(E reference) {
            iterator.set(reference);
            if (length > 1) {   // Unable to compare against replaced object.
                removeCharacteristics(SORTED | DISTINCT);
            }
        }

        @Override
        public void add(E reference) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator());
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        return new ArrayIteratorImpl(elements.subList(offset, offset + length).listIterator(index));
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    @Override
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = -1498144741859632427L;

        transient int offset;
        final int length;
        final int characteristics;
        final Comparator<? super E> comparator;
        transient List<E> elements;

        SerializationProxy(MutableListArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            if (offset == 0 && length == elements.size()) {
                s.writeObject(elements);
            } else {
                s.writeObject(elements.subList(offset, offset + length));
            }
        }

        @SuppressWarnings("unchecked")
        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (List<E>) s.readObject();
        }

        private Object readResolve() {
            return new MutableListArrayImpl<>(elements, 0, length, characteristics, comparator);
        }
    }
}
