/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Comparator;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.function.Consumer;

import static java.lang.System.arraycopy;
import static java.lang.reflect.Array.newInstance;
import static java.util.Arrays.copyOfRange;
import static java.util.Objects.requireNonNull;

/**
 * <p>Immutable array of object references.</p>
 *
 * <p>Object arrays are constructed by the static interface methods of {@code ObjectArray}. The {@code copyOf} method
 * returns a new array representing a shallow copy of the array argument. The
 * {@code unsafeValueOf} methods wrap a new {@code ObjectArray} around a native array. It is imperative that this native
 * array is not modified outside the scope of {@code ObjectArray}, as this would break the immutable contract. The
 * {@code unsafeValueOf} method is used to prevent unnecessary array copies. It is the caller's responsibility not to
 * violate the contract.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This
 * means that this implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. The set of characteristics of immutable
 * arrays are constant.</p>
 *
 * @author Leon van Zantvoort
 */
@Immutable
class ObjectArrayImpl<E> extends AbstractArray<E> implements NativeArray<E>, ObjectArray<E>, Serializable {

    /** Backing array. */
    protected transient Object[] elements;

    /** Comparator that is used for sorting array. */
    protected Comparator<? super E> comparator;

    /**
     * Package visible constructor, to be used by static factory methods only.
     *
     * @param array backing array.
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> is used.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    ObjectArrayImpl(@Nonnull Object[] array, int offset, int length, int characteristics,
                    @Nullable Comparator<? super E> comparator, boolean copy) {
        super(copy ? 0 : offset, length, characteristics);
        requireNonNull(array);
        if (length > array.length - offset) {
            throw new IndexOutOfBoundsException();
        }
        if (copy) {
            this.elements = (Object[]) java.lang.reflect.Array.newInstance(array.getClass().getComponentType(), length);
            arraycopy(array, offset, elements, 0, length);
        } else {
            this.elements = array;
        }
        this.length = length;
        this.comparator = comparator;
        if (length == 0) {
            this.characteristics |= NONNULL;
        }
    }

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    @SuppressWarnings("unchecked")
    @Override
    public final Class<? super E> getComponentType() {
        return (Class<Object>) elements.getClass().getComponentType();
    }

    /**
     * Tells whether or not this array is backed by an accessible array.
     *
     * <p>If this method returns {@code true} then the {@link #nativeArray()} method may safely be invoked.</p>
     *
     * @return {@code true} if, and only if, this array is backed by an accessible array.
     */
    @Override
    public boolean hasArray() {
        return true;
    }

    /**
     * <p>Returns a native array containing all elements of this array.</p>
     *
     * @return native array.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    @Override
    public final E[] nativeArray() {
        return (E[]) elements;
    }

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @Override
    public final @CheckForNull Comparator<? super E> getComparator() {
        return comparator;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    protected int filterCharacteristics(int characteristics) {
        return characteristics & ~(MUTABLE | BOUNDED | UNBOUNDED);
    }

    /**
     * Returns the object found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @SuppressWarnings("unchecked")
    @Override
    public @Nullable E get(int index) {
        rangeCheck(index);
        return (E) elements[offset + index];
    }

    /**
     * Returns the object found at {@code [offset + index]} of the backing array. This method allows for fast
     * failure detection.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     * @throws NullPointerException if null reference is found at index.
     */
    @SuppressWarnings("unchecked")
    @Override
    public @Nonnull E getNonnull(int index) {
        rangeCheck(index);
        E value = (E) elements[offset + index];
        requireNonNull(value, "Assertion error, nonnull reference expected.");
        return value;
    }

    /**
     * Returns {@code true} if this array contains the specified object. This method supports boxed primitives. This
     * means that if this array represents a primitive array, this method returns {@code true} if this array contains
     * the unboxed version of specified object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object object whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified object.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public boolean contains(Object object) {
        if (hasCharacteristics(SORTED)) {
            Comparator<? super E> comparator = getComparator();
            // No NullPointerException in case of natural ordering.
            return !(object == null && comparator == null) && Arrays.binarySearch(this, (E) object, comparator) >= 0;
        }
        final Object[] a = elements;
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            if (Objects.equals(a[index], object)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns {@code true} if this array contains all of the elements in the specified {@code array}. Primitives and
     * their object counterparts are regarded equal in this respect.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array is sorted, otherwise all elements are
     * traversed until the objects in specified array are found.</p>
     *
     * @param array array to be checked for containment in this array.
     * @return {@code true} if this array contains all of the elements in the specified array.
     * @throws NullPointerException if the specified array is null, or if the specified array contains null elements and
     * array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with elements.
     */
    @Override
    public boolean containsAll(@Nonnull Array<?> array) {
        // Fastest: direct array access.
        Object[] backingArray = NativeArrays.array(array);
        if (backingArray != null) {
            for (int index = array.offset(), toIndex = index + array.length(); index < toIndex; index++) {
                if (!contains(backingArray[index])) {
                    return false;
                }
            }
            return true;
        }
        // Alternative.
        for (Object o : array) {
            if (!contains(o)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns the index of the first occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public int indexOf(Object object) {
        if (hasCharacteristics(SORTED)) {
            Comparator<? super E> comparator = getComparator();
            if (object == null && comparator == null) {  // No NullPointerException in case of natural ordering.
                return -1;
            }
            int index = Arrays.binarySearch(this, (E) object, comparator);
            if (index >= 0) {
                return Arrays.binarySearchFirst(this, index, comparator);
            }
            return -1;
        }
        final Object[] a = elements;
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            if (Objects.equals(a[index], object)) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified object is found.</p>
     *
     * @param object object whose highest index in this array is to be returned.
     * @return highest index of object in array or {@code -1} if value is not found.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    @SuppressWarnings("unchecked")
    @Override
    public int lastIndexOf(Object object) {
        if (hasCharacteristics(SORTED)) {
            Comparator<? super E> comparator = getComparator();
            if (object == null && comparator == null) {  // No NullPointerException in case of natural ordering.
                return -1;
            }
            int index = Arrays.binarySearch(this, (E) object, comparator);
            if (index >= 0) {
                return Arrays.binarySearchLast(this, index, comparator);
            }
            return -1;
        }
        final Object[] a = elements;
        for (int toIndex = offset + length, index = toIndex - 1; index >= offset; index--) {
            if (Objects.equals(a[index], object)) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @SuppressWarnings("unchecked")
    @Override
    public Object[] toArray() {
        return copyOfRange(elements, offset, offset + length);
    }

    /**
     * Returns an array containing all of the reference of {@code this} array. If the references fit in the
     * specified array, it is returned therein. Otherwise, a new array of the same object type is allocated.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T> T[] toArray(@Nonnull T[] array) {
        T[] a = array.length >= length ? array : (T[]) newInstance(array.getClass().getComponentType(), length);
        arraycopy(elements, offset, a, 0, length);
        return a;
    }

    /**
     * Stores elements from this array into the specified array.
     *
     * @param <T> custom element type.
     * @param array the array into which the elements of this array are to be stored.
     * @param offset the offset within the specified array of the first element to be written.
     * @param length the number of elements to be stored. Must be smaller than length of this array.
     * @throws IndexOutOfBoundsException if offset and length parameters are invalid.
     */
    @Override
    public <T> void toArray(@Nonnull T[] array, int offset, int length) {
        arraycopy(elements, this.offset, array, offset, length);
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     */
    @SuppressWarnings("unchecked")
    @Override
    public void forEach(@Nonnull Consumer<? super E> consumer) {
        requireNonNull(consumer);
        final Object[] a = elements;
        for (int index = offset, toIndex = index + length; index < toIndex; index++) {
            consumer.accept((E) a[index]);
        }
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @SuppressWarnings("unchecked")
    @Override
    public ObjectArray<E> offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new ObjectArrayImpl<>(elements, newOffset, newLength, characteristics, comparator, false);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @SuppressWarnings("unchecked")
    @Override
    public ObjectArray<E> length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new ObjectArrayImpl<>(elements, offset, length, characteristics, comparator, false);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    @SuppressWarnings("unchecked")
    public ObjectArray<E> subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new ObjectArrayImpl<>(elements, newOffset, newLength, characteristics, comparator, false);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator<E> {
        int index;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @SuppressWarnings("unchecked")
        @Override
        public E next() {
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i + 1;
            return (E) elements[i];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @SuppressWarnings("unchecked")
        @Override
        public E previous() {
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i;
            return (E) elements[i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @Override
        public void set(E reference) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void add(E reference) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator<E> iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator<E> iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(index);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public String toString() {
        int iMax = offset + length - 1;
        if (iMax - offset == -1) {
            return "[]";
        }
        StringBuilder b = new StringBuilder();
        b.append('[');
        for (int i = offset; ; i++) {
            b.append(elements[i]);
            if (i == iMax) {
                return b.append(']').toString();
            }
            b.append(", ");
        }
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public int hashCode() {
        int result = 1;
        int to = offset + length;
        for (int i = offset; i < to; i++) {
            Object element = elements[i];
            result = 31 * result + (element == null ? 0 : element.hashCode());
        }
        return result;
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj instanceof Array) {
            Array<?> array = (Array<?>) obj;
            if (length != array.length()) {
                return false;
            }
            // Fastest: direct array access.
            Object[] backingArray = NativeArrays.array(array);
            if (backingArray != null) {
                final int o = array.offset();
                for (int i = 0; i < length; i++) {
                    if (!Objects.equals(elements[offset + i], backingArray[o + i])) {
                        return false;
                    }
                }
                return true;
            }
            // Alternative.
            return Arrays.equals(this, array);
        }
        return false;
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    Object writeReplace() {
        return new SerializationProxy<>(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy<E> implements Serializable {
        static final long serialVersionUID = 8589659850953249192L;

        transient int offset;
        final int length;
        final int characteristics;
        final Comparator<? super E> comparator;
        final Class<?> cls;
        transient Object[] elements;

        SerializationProxy(ObjectArrayImpl<E> array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.comparator = array.comparator;
            this.elements = array.elements;
            this.cls = this.elements.getClass().getComponentType();
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int index = offset, toIndex = index + length; index < toIndex; index++) {
                s.writeObject(elements[index]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = (Object[]) newInstance(cls, length);
            for (int index = 0; index < length; index++) {
                elements[index] = s.readObject();
            }
        }

        @SuppressWarnings("unchecked")
        private Object readResolve() {
            return new ObjectArrayImpl<>(elements, 0, length, characteristics, comparator, false);
        }
    }
}