/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import org.squarebrackets.function.CharPredicate;

import javax.annotation.Nonnull;

import static java.util.Objects.requireNonNull;

/**
 * <p>Capacity unbounded array of char primitives.</p>
 *
 * <p>Unbounded char arrays are constructed by the static interface methods of this interface. The {@code newInstance}
 * method returns a new empty array. The {@code copyOf} method returns a new array representing a deep copy of the
 * array argument. The {@code valueOf} methods wrap a new {@code UnboundedByteArray} around a native array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Unbounded primitive arrays, such as {@code UnboundedCharArray}, always report {@code UNBOUNDED}, {@code MUTABLE}
 * and {@code NONNULL}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 * 
 * @author Leon van Zantvoort
 */
public interface UnboundedCharArray extends DynamicCharArray, UnboundedArray<Character> {

    /**
     * Creates an empty unbounded array with an initial capacity of ten.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code UNBOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @return a new unbounded array instance.
     */
    static UnboundedCharArray newInstance() {
        return newInstance(10);
    }

    /**
     * Creates an empty array with the specified initial capacity.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code UNBOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param initialCapacity the initial capacity of the array.
     * @throws IllegalArgumentException if the specified initial capacity is negative.
     * @return a new unbounded array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static UnboundedCharArray newInstance(int initialCapacity) {
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal capacity: " + initialCapacity);
        }
        return valueOf(new char[initialCapacity], 0, 0);
    }

    /**
     * Creates a capacity unbounded copy of the specified {@code array}, backed by a native array. If the specified
     * {@code array} is mutable, changes to the mutable array are not reflected by this unbounded copy.
     *
     * <p>The specified {@code array} can be an array of {@code char} primitives or {@code Character} objects. The
     * specified {@code array} must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code UNBOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}.</li>
     * </ul>
     *
     * @param array array to be copied.
     * @return a capacity unbounded array of char values.
     * @throws NullPointerException if the specified array is null or if the array contains null elements.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static UnboundedCharArray copyOf(@Nonnull Array<Character> array) {
        char[] backingArray = NativeArrays.charArray(array);
        if (backingArray != null) {
            return Arrays.newUnboundedArray(backingArray, array.offset(), array.length(), array.characteristics(), true);
        }
        return Arrays.newUnboundedArray(NativeArrays.toCharArray(array), 0, array.length(), array.characteristics(), false);
    }

    /**
     * Creates an unbounded array of the specified {@code array}. The specified {@code array} is used as backing array,
     * but is replaced by a larger array when more capacity is needed.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code UNBOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @return an unbounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static UnboundedCharArray valueOf(@Nonnull char... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates an unbounded array of the specified {@code array}. The specified {@code array} is used as backing array,
     * but is replaced by a larger array when more capacity is needed.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code UNBOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return an unbounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static UnboundedCharArray valueOf(@Nonnull char[] array, int offset, int length) {
        return Arrays.newUnboundedArray(array, offset, length, 0, false);
    }

    /**
     * Inserts the specified char value at the specified {@code index}, moving elements at and after {@code index} to the
     * right. Length of {@code this} array is increased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect this array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if this array reports {@code SORTED} prior to this call, {@code SORTED} is reported after this
     *   invocation if the specified {@code object} does not break the sorted order of this array;</li>
     *   <li>if this array reports {@code SORTED} and {@code DISTINCT} prior to this call, {@code DISTINCT} is
     *   reported after this invocation if the specified {@code object} was not already present in this array.</li>
     * </ul>
     *
     * @param index insertion index, relative to the offset of the array.
     * @param value value to be inserted.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    void addChar(int index, char value);

    /**
     * Removes element at specified {@code index}, moving all elements after {@code index} to the left. Length of
     * {@code this} array is decreased by one.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param index removal index, relative to the offset of the array.
     * @return element that is removed.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    char charRemove(int index);

    /**
     * Removes the first occurrence of the specified value from this array if it is present. If this array does not
     * contain the value, it is unchanged.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics to lookup the element to be removed, otherwise all elements are traversed until the specified
     * value is found.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>if the array contains one element after this invocation, {@code SORTED} and {@code DISTINCT} are reported.
     * </ul>
     *
     * @param value value to be removed from this array.
     * @return {@code true} if this array contained the specified value.
     */
    boolean removeChar(char value);

    /**
     * Removes all of the elements of this array's range that satisfy the given predicate. Errors or runtime exceptions
     * thrown by the predicate are relayed to the caller.
     *
     * <p>The default implementation traverses all elements of the array using its {@link #iterator}. Each matching
     * element is removed using {@link java.util.Iterator#remove()}.</p>
     *
     * @param filter a predicate which returns {@code true} for elements to be removed.
     * @return {@code true} if any elements were removed.
     * @throws NullPointerException if the specified filter is null.
     */
    default boolean removeIfChar(@Nonnull CharPredicate filter) {
        requireNonNull(filter);
        boolean removed = false;
        final ArrayIterator.OfChar each = iterator();
        while (each.hasNext()) {
            if (filter.test(each.nextChar())) {
                each.remove();
                removed = true;
            }
        }
        return removed;
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    UnboundedCharArray offset(int offset);     // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    UnboundedCharArray length(int length);     // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    UnboundedCharArray subArray(int fromIndex, int toIndex);   // More specific return type.
}
