/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.util.Comparator;

import static java.util.Objects.requireNonNull;

/**
 * <p>Capacity unbounded array of object references.</p>
 *
 * <p>Unbounded object arrays are constructed by the static interface methods of this interface. The {@code newInstance}
 * method returns a new empty array. The {@code copyOf} method returns a new array representing a shallow copy of the
 * array argument. The {@code valueOf} methods wrap a new {@code UnboundedObjectArray} around a native array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Unbounded arrays, such as {@code UnboundedObjectArray}, always report {@code UNBOUNDED} and {@code MUTABLE}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 * 
 * <p>Dynamic object arrays may eliminate obsolete references to allow the garbage collector to reclaim memory.
 * This is done by setting the elements in 'reserved space' to {@code null}. This feature can be disabled by
 * calling {@code setClearReservedElements(false)}.</p>
 *
 * @param <E> element type.
 * @author Leon van Zantvoort
 */
public interface UnboundedObjectArray<E> extends DynamicObjectArray<E>, UnboundedArray<E> {

    /**
     * Creates an empty unbounded array with an initial capacity of ten. The backing array will be of type Object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code UNBOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @return a new unbounded array instance.
     */
    @SuppressWarnings("unchecked")
    static <E> UnboundedObjectArray<E> newInstance() {
        return (UnboundedObjectArray<E>) newInstance(Object.class, 10);
    }

    /**
     * Creates an empty unbounded array with an initial capacity of ten. The backing array will be of type {@code cls}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code UNBOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param cls component type.
     * @return a new unbounded array instance.
     * @throws NullPointerException if the specified class is null.
     */
    static <E> UnboundedObjectArray<E> newInstance(@Nonnull Class<E> cls) {
        return newInstance(cls, 10);
    }

    /**
     * Creates an empty array with the specified initial capacity. The backing array will be of type Object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code UNBOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param initialCapacity the initial capacity of the array.
     * @param <E> element type.
     * @throws IllegalArgumentException if the specified initial capacity is negative.
     * @return a new unbounded array.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    static <E> UnboundedObjectArray<E> newInstance(int initialCapacity) {
        return (UnboundedObjectArray<E>) newInstance(Object.class, initialCapacity);
    }

    /**
     * Creates an empty array with the specified initial capacity. The backing array will be of type {@code cls}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code UNBOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param initialCapacity the initial capacity of the array.
     * @param <E> element type.
     * @param cls component type.
     * @throws IllegalArgumentException if the specified initial capacity is negative.
     * @return a new unbounded array.
     * @throws NullPointerException if the specified class is null.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    static <E> UnboundedObjectArray<E> newInstance(@Nonnull Class<E> cls, int initialCapacity) {
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal capacity: " + initialCapacity);
        }
        return valueOf((E[]) java.lang.reflect.Array.newInstance(requireNonNull(cls), initialCapacity), 0, 0);
    }

    /**
     * Creates a capacity unbounded copy of the specified {@code array}, backed by native array. If the specified
     * {@code array} is mutable, changes to the mutable array are not reflected by this unbounded copy.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE} and {@code UNBOUNDED};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param array array to be copied.
     * @return a capacity unbounded array of objects.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    static <E> UnboundedObjectArray<E> copyOf(@Nonnull Array<E> array) {
        E[] backingArray = (E[]) NativeArrays.array(array);
        if (backingArray != null) {
            return Arrays.newUnboundedArray(backingArray, array.offset(), array.length(), array.characteristics(),
                    array.getComparator(), true);
        }
        return Arrays.newUnboundedArray((E[]) NativeArrays.toArray(array), 0, array.length(), array.characteristics(),
                array.getComparator(), false);
    }

    /**
     * Creates an unbounded array of the specified {@code array}. The specified {@code array} is used as backing array,
     * but is replaced by a larger array when more capacity is needed.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE} and {@code UNBOUNDED};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param <E> element type.
     * @return an unbounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @SafeVarargs
    static <E> UnboundedObjectArray<E> valueOf(@Nonnull E... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates an unbounded array of the specified {@code array}. The specified {@code array} is used as backing array,
     * but is replaced by a larger array when more capacity is needed.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE} and {@code UNBOUNDED};</li>
     *   <li>{@code NONNULL} if the specified {@code length} is zero;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param <E> element type.
     * @return an unbounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static <E> UnboundedObjectArray<E> valueOf(@Nonnull E[] array, int offset, int length) {
        return Arrays.newUnboundedArray(array, offset, length, 0, null, false);
    }

    /**
     * Specifies if reserved elements need to be reset to default values. This flag is set to {@code true} by default
     * for object arrays.
     *
     * @param flag {@code true} if reserved need to be reset, {@code false} otherwise.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    void setClearReservedElements(boolean flag);

    /**
     * Removes all duplicates from this array. The order of elements may be altered as result of this call. The
     * {@code comparator} can be retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} and {@code DISTINCT} are reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> should be used.
     * @throws IllegalArgumentException (optional) if the comparator is found to violate the
     * {@link java.util.Comparator} contract.
     * @throws NullPointerException if this array contains null elements and comparator does not permit null arguments.
     * @throws ClassCastException if this array contains non {@code Comparable} elements or comparator not compatible
     * with elements.
     */
    void distinct(Comparator<? super E> comparator);

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    UnboundedObjectArray<E> offset(int offset);    // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    UnboundedObjectArray<E> length(int length);   // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    UnboundedObjectArray<E> subArray(int fromIndex, int toIndex); // More specific return type.
}
