/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets.function;

import static java.util.Objects.requireNonNull;

/**
 * An operation which accepts a single byte argument and returns no result.
 *
 * @author Leon van Zantvoort
 */
@FunctionalInterface
public interface ByteConsumer {

    /**
     * Accept an input value.
     *
     * @param value the input value
     */
    void accept(byte value);

/**
 * Returns an {@code ByteConsumer} which performs, in sequence, the operation represented by this object followed 
 * by the operation represented by another {@code ByteConsumer}.
 *
 * <p>Any exceptions thrown by either {@code accept} method are relayed to the caller; if performing this operation
 * throws an exception, the other operation will not be performed.
 *
 * @param other an ByteConsumer which will be chained after this ByteConsumer.
 * @return an ByteConsumer which performs in sequence the {@code accept} method of this ByteConsumer and the
 * {@code accept} method of the specified ByteConsumer operation.
 * @throws NullPointerException if other is null.
 */
default ByteConsumer chain(ByteConsumer other) {
    requireNonNull(other);
    return (byte value) -> { accept(value); other.accept(value); };
}
}
